"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Session = void 0;
const chokidar = require("chokidar");
const path = require("path");
const getSessionConfiguration_1 = require("./getSessionConfiguration");
const write_1 = require("../util/write");
const parseCSS_1 = require("./parseCSS");
const extractPluginResult_1 = require("./extractPluginResult");
const fs_1 = require("../util/fs");
const generateScript_1 = require("./generateScript");
const waitForInitialScanCompletion_1 = require("./waitForInitialScanCompletion");
const minifyScripts_1 = require("../minifier/minifyScripts");
const createExposedPromise_1 = require("../util/createExposedPromise");
const minifyScriptsForCSS_1 = require("../minifier/minifyScriptsForCSS");
class Session {
    constructor(parameters = {}) {
        this.configuration = getSessionConfiguration_1.getSessionConfiguration(parameters);
        this.processedFiles = new Set();
        this.initialTask = null;
        this.tasks = new Set();
    }
    get helperPath() {
        const srcDirectory = path.join(__dirname, '..', 'helper');
        const name = this.configuration.output.type === 'script' ? 'default' : 'noop';
        return path.join(srcDirectory, `${name}${this.configuration.ext}`);
    }
    async start() {
        await this.outputHelperScript();
        await this.startWatcher();
    }
    async stop() {
        await this.stopWatcher();
    }
    async outputHelperScript() {
        if (this.configuration.output.type === 'script') {
            await fs_1.copyFile(this.helperPath, this.configuration.output.path);
        }
    }
    async processCSS(filePath) {
        if (0 < this.tasks.size) {
            await this.waitCurrentTasks();
        }
        const exposedPromise = this.createExposedPromise();
        const { configuration } = this;
        const postcssResult = await parseCSS_1.parseCSS({
            plugins: configuration.postcssPlugins,
            options: configuration.postcssOptions,
            file: filePath,
        })
            .catch((error) => {
            exposedPromise.resolve();
            throw error;
        });
        const dest = path.join(`${filePath}${configuration.ext}`);
        this.processedFiles.add(filePath);
        const { output } = configuration;
        const code = generateScript_1.generateScript({
            output: dest,
            helper: output.type === 'css' ? this.helperPath : output.path,
            result: extractPluginResult_1.extractPluginResult(postcssResult),
            root: postcssResult.root,
            cssKey: this.configuration.cssKey,
        });
        exposedPromise.resolve();
        return { dest, code };
    }
    async minifyScripts() {
        const files = [...this.processedFiles].map((file) => `${file}${this.configuration.ext}`);
        const { cssKey, output } = this.configuration;
        if (output.type === 'css') {
            await minifyScriptsForCSS_1.minifyScriptsForCSS({ files, cssKey, dest: output.path });
        }
        else {
            await minifyScripts_1.minifyScripts({ files, cssKey, dest: output.path });
        }
    }
    async waitCurrentTasks() {
        await Promise.all([...this.tasks]);
    }
    createExposedPromise() {
        const exposedPromise = createExposedPromise_1.createExposedPromise();
        this.tasks.add(exposedPromise.promise);
        const removeTask = () => this.tasks.delete(exposedPromise.promise);
        exposedPromise.promise.then(removeTask).catch(removeTask);
        return exposedPromise;
    }
    async startWatcher() {
        await this.stopWatcher();
        this.initialTask = [];
        this.log(`watching: ${this.configuration.path.join(', ')}`);
        const watcher = this.watcher = chokidar.watch(this.configuration.path.slice(), this.configuration.chokidar)
            .on('error', this.onError.bind(this))
            .on('all', (eventName, file, stats) => {
            this.onFileEvent(eventName, file, stats)
                .catch((error) => {
                watcher.emit('error', error);
            });
        });
        await waitForInitialScanCompletion_1.waitForInitialScanCompletion(this.watcher);
        await Promise.all(this.initialTask);
        this.initialTask = null;
        if (!this.configuration.watch) {
            await this.minifyScripts();
            await this.stop();
        }
    }
    log(...messages) {
        write_1.write(this.configuration.stdout, messages);
    }
    logError(...messages) {
        write_1.write(this.configuration.stderr, messages);
    }
    async stopWatcher() {
        if (this.watcher) {
            await this.watcher.close();
            delete this.watcher;
        }
        await Promise.resolve(null);
    }
    onError(error) {
        this.logError(error);
    }
    async onFileEvent(eventName, file, stats) {
        this.log(`[${eventName}] ${file}`);
        switch (eventName) {
            case 'add': {
                const promise = this.onAdd(file, stats);
                if (this.initialTask) {
                    this.initialTask.push(promise);
                }
                await promise;
                break;
            }
            case 'change':
                await this.onChange(file, stats);
                break;
            case 'unlink':
                await this.onUnlink(file);
                break;
            default:
        }
    }
    async onAdd(file, stats) {
        await this.onChange(file, stats);
    }
    async onChange(file, stats) {
        if (!stats) {
            throw new Error(`no stats is given for ${file}.`);
        }
        if (!stats.isFile()) {
            throw new Error(`${file} is not a file.`);
        }
        const { dest, code } = await this.processCSS(file);
        await fs_1.writeFile(dest, code, this.configuration.stdout);
    }
    async onUnlink(file) {
        const outputPath = path.join(`${file}${this.configuration.ext}`);
        this.processedFiles.delete(file);
        await fs_1.deleteFile(outputPath, this.configuration.stdout);
    }
}
exports.Session = Session;